from typing import Union
from datetime import (
    date,
    datetime
)

import pandas
import altair
from pandas._typing import FilePathOrBuffer

from conjurer.logic.eda import check
from conjurer.logic.eda.load import (
    pandas_csv,
    df_dict_loader
)
from conjurer.logic.eda.vis import (
    histogram,
    scatter
)


StrOrList = Union[str, list]
Orderable = Union[int, float, datetime, date]


def read_csv(buffer_or_filepath: FilePathOrBuffer, **kwargs) -> pandas.DataFrame:
    """
    Load CSV as pandas.DataFrame with Int64 / datetime64 column inference
    Args:
        buffer_or_filepath: 1st argument for pandas.read_csv
        **kwargs: Arbitrary keyword arguments for pandas.read_csv

    Returns:
        pandas.DataFrame
    """
    return pandas_csv.read_csv(buffer_or_filepath, **kwargs)


def check_stats(df: pandas.DataFrame, skip_histogram: bool = False) -> pandas.DataFrame:
    """
    Calculate basic statistics for pandas.DataFrame
    Args:
        df (pandas.DataFrame): Data frame you want to calculate statistics
        skip_histogram (bool, optional): If True, function does not generate histogram

    Returns:
        pandas.DataFrame: Data frame of statistics for each column

        Each row represents statistics of each column in `df`
    """
    return check.check_stats(df, skip_histogram)


def get_unique_values(df: pandas.DataFrame, columns: StrOrList) -> set:
    """
    Get unique values (not null) for column(s) in pandas.DataFrame as set
    Args:
        df (pandas.DataFrame): Data frame which stores data
        columns (str or list of str): Column name(s) you want to confirm unique values

    Returns:
        set: unique values (only values which does NOT include any nulls)

    Notes:
        like https://github.com/numpy/numpy/issues/9358, "unique" concept for null is ambiguous,
        values which includes null are all eliminated from returned value
    """
    return check.get_unique_values(df, columns)


def get_columns_in_dfs(df_list: list, name_list: list) -> pandas.DataFrame:
    """
    Summarize df names and column name in multiple pandas.DataFrame
    Args:
        df_list (list of pandas.DataFrame): List of data frames
        name_list (list of str): Name of each data frame

    Returns:
        pandas.DataFrame: with columns "table_name" and "column_name"
    """
    return check.get_columns_in_dfs(df_list, name_list)


def get_fk_coverage(
    fk_df: pandas.DataFrame, k_df: pandas.DataFrame, fk_columns: StrOrList, k_columns: StrOrList,
    do_print: bool = True) -> float:
    """
    Check how many keys in `fk_df.fk_columns` exists in `k_df.k_columns`
    Args:
        fk_df (pandas.DataFrame): Data frame which has FK columns
        k_df (pandas.DataFrame): Data frame which has K columns
        fk_columns (str or list of str): FK columns
        k_columns (str or list of str): K columns
        do_print (bool): If True, print results

    Returns:
        float: ratio of records in `fk_df.fk_columns` which has records in `k_df.k_columns`
    """
    return check.get_fk_coverage(fk_df, k_df, fk_columns, k_columns, do_print)


def plot_histogram(series: pandas.Series, num_bins: int = 50, normalize: bool = False,
                   minv: Orderable = None, maxv: Orderable = None) -> altair.Chart:
    """
    Plot histogram with altair
    Args:
        series (pandas.Series): Values you want to plot histogram
        num_bins (int, optional): Default=50. The number of bins used in the histogram
        normalize (bool, optional): Default=False. If True, vertical axis is ratio of records, else the number of records in each bin
        minv (optional): Default=Minimum value in array. Minimum value to generate bins
        maxv (optional): Default=Maximum value in array. Maximum value to generate bins

    Returns:
       altair.Chart
    """
    return histogram.plot_histogram(series, num_bins, normalize, minv, maxv)


def plot_scatter(df: pandas.DataFrame, column_x: str, column_y: str, **kwargs) -> altair.Chart:
    """
    Plot scatter plot with altair. If sample size is too large, plot heatmap instead
    Args:
        df(pandas.DataFrame): Data frame which stores all the data used in the chart
        column_x(str): Column name for x value
        column_y(str): Column name for y value
        num_bins_x(int): The number of bins (used for heatmap)
        num_bins_y(int): The number of bins (used for heatmap)
        xmin(numeric or timestamp): Minimum X value of plot
        xmax(numeric or timestamp): Maximum X value of plot
        ymin(numeric or timestamp): Minimum Y value of plot
        ymax(numeric or timestamp): Maximum Y value of plot

    Returns:
        altair.Chart
    """
    return scatter.plot_scatter(df, column_x, column_y, **kwargs)


def plot_heatmap(df: pandas.DataFrame, column_x: str, column_y: str, num_bins_x: int = 50, num_bins_y: int = 50,
                 xmin: Orderable = None, xmax: Orderable = None, ymin: Orderable = None, ymax: Orderable = None) \
        -> altair.Chart:
    """
    Plot heatmap with altair.
    Args:
        df(pandas.DataFrame): Data frame which stores all the data used in the chart
        column_x(str): Column name for x value
        column_y(str): Column name for y value
        num_bins_x(int): The number of bins
        num_bins_y(int): The number of bins
        xmin(numeric or timestamp): Minimum X value of plot
        xmax(numeric or timestamp): Maximum X value of plot
        ymin(numeric or timestamp): Minimum Y value of plot
        ymax(numeric or timestamp): Maximum Y value of plot

    Returns:
        altair.Chart
    """
    return scatter.plot_heatmap(df, column_x, column_y, num_bins_x, num_bins_y, xmin, xmax, ymin, ymax)


DfDictLoader = df_dict_loader.DfDictLoader



from numpy import random
import pandas


def get_input_df(sample_size):
    return pandas.DataFrame({
        **{
            "id1": range(sample_size),
            "id2": range(10000, 10000+sample_size),
            "target_cl": random.choice([1, 0], sample_size),
            "target_rg": random.normal(0, 1, sample_size),
        },
        **{
            "column{}".format(i): random.normal(0, 1, sample_size)
            for i in range(3)
        },
        **{
            "column{}".format(i): random.choice([1, 0], sample_size)
            for i in range(3, 6)
        }
    })


def get_target_column(is_cl):
    return "target_cl" if is_cl else "target_rg"


def get_xy(sample_size, for_cl):
    input_df = get_input_df(sample_size)
    feature_columns = ["column{}".format(i) for i in range(6)]
    target_column = "target_cl" if for_cl else "target_rg"
    return input_df[feature_columns].values, input_df[target_column].values


def assert_prediction(model, df_test, is_cl):
    pred_df = model.predict(df_test)
    expected_columns = ["score", "id1", "id2", "target_cl", "target_rg"]
    if is_cl:
        expected_columns.insert(1, "predicted_class")
    assert list(pred_df.columns) == expected_columns
    assert len(pred_df) == 10



from conjurer import ml
from .tuner_config.utils import make_pipeline
from .tuner_config import (
    lightgbm,
    xgboost,
    random_forest,
    linear_model
)

default_scoring = {
    "cl": "roc_auc",
    "rg": "neg_mean_squared_error",
    "mcl": "accuracy"
}


def get_cv(ml_type, problem_type, scoring=None, search_type="random", n_iter=20, **kwargs):
    """
    Get RandomizedSearchCV or GridSearchCV object for tuning
    Args:
        ml_type (str): One of "lightgbm", "xgboost", "random_forest", "linear_model"
        problem_type (str): One of "cl", "rg", "mcl"
        scoring (str or scorer): `scoring` for RandomizedSearchCV / GridSearchCV
        search_type (str): Return RandomizedSearchCV if "random", GridSearchCV if "grid"
        param_dict (dict): param_distribution or param_grid for CV object
        n_iter (int): `n_iter` for RandomizedSearchCV
        kwargs: keyword arguments for CV object

    Returns:
        conjurer.ml.RandomizedSearchCV or conjurer.ml.GridSearchCV
    """
    _validate_attr(ml_type, problem_type, search_type)
    config = getattr(globals()[ml_type], "config")
    scoring = scoring or default_scoring[problem_type]
    ml_estimator = config.estimator_dict[problem_type]
    estimator = make_pipeline.add_default_preprocessing("ml", ml_estimator)
    param_dict = _get_param_dict(problem_type, search_type, "ml", config.parameters_dict)
    cv_class = ml.RandomizedSearchCV if search_type == "random" else ml.GridSearchCV
    if search_type == "random":
        kwargs["n_iter"] = n_iter
    return cv_class(estimator, param_dict, scoring=scoring, **kwargs)


def _validate_attr(ml_type, problem_type, search_type):
    assert ml_type in ["lightgbm", "xgboost", "linear_model", "random_forest"], \
        "`ml_type` should be one of lightgbm, xgboost, linear_model, or random_forest"
    assert problem_type in ["cl", "rg", "mcl"], "`problem_type` should be one of cl, rg, or mcl"
    assert search_type in ["random", "grid"], "`search_type` should be one of random or grid"


def _get_param_dict(problem_type, cv_type, name, parameters_dict):
    ml_params = parameters_dict[cv_type][problem_type]
    return make_pipeline.convert_param_dict(name, ml_params)



import math
import copy
import logging

import numpy
import pandas
from pandas.api import types


logger = logging.getLogger(__name__)


def calc_column_stat(df):
    df_size = len(df)
    output_column_names = [
        "column_name", "dtype", "min", "max", "mean", "std", "ratio_na", 
        "ratio_zero", "unique_count", "is_unique"
    ]
    def _calc_stat(column, df_size):
        unique_count = len(get_unique_values(df, column))
        logger.info("...calculating {} (dtype: {})".format(column, df.dtypes[column]))
        return pandas.DataFrame(
            {
                "column_name": [column],
                "dtype": df.dtypes[column],
                "min": [df[column].min()] if _orderable(df.dtypes[column]) else [pandas.NA],
                "max": [df[column].max()] if _orderable(df.dtypes[column]) else [pandas.NA],
                "mean": [df[column].mean()] if types.is_numeric_dtype(df.dtypes[column]) else [pandas.NA],
                "std": [df[column].std()] if types.is_numeric_dtype(df.dtypes[column]) else [pandas.NA],
                "ratio_na": [pandas.isnull(df[column]).sum() / float(df_size)],
                "ratio_zero": [len(df[df[column]==0]) / float(df_size)]\
                    if types.is_numeric_dtype(df.dtypes[column]) else [pandas.NA],
                "unique_count": [unique_count],
                "is_unique": [unique_count == df_size]
            },
            columns=output_column_names
        )

    return pandas.concat([_calc_stat(c, df_size) for c in df.columns], axis=0)


def get_unique_values(df, columns):
    df_tmp = df[columns].dropna()
    if isinstance(df_tmp, pandas.Series):
        return set([v for v in df_tmp.values])
    else:
        return set([tuple(v) for v in df_tmp.values])


def calculate_percentiles_for_df(df, column_list, ratio_list):
    value_dic = {
        **{
            "name": ["min"] +
                     ["{0:.2%}-percentile".format(ratio) for ratio in ratio_list]
                     + ["max"]
        },
        **{
            c: calculate_percentiles(df[column].values, ratio_list)
            for c in column_list
        }
    }
    return pandas.DataFrame(value_dic, columns=["name"] + column_list).set_index("name")


def calculate_percentiles(array, ratio_list):
    n_record0 = array.shape[0]
    valid_array = array[numpy.isfinite(array)]
    n_record = valid_array.shape[0]
    if n_record0 - n_record > 0:
        logger.info("{0} records have missing values (out of {1} records)".format(n_record0 - n_record, n_record0))
    sorted_array = numpy.sort(valid_array)
    ind_list = [_get_ind(n_record, ratio) for ratio in ratio_list]
    return numpy.array([sorted_array[ind] for ind in ind_list])


def _get_ind(n_record, ratio):
    return 0 if ratio == 0 else int(math.ceil(n_record * ratio)) - 1


def _orderable(dtype):
    if types.is_numeric_dtype(dtype) or types.is_datetime64_any_dtype(dtype):
        return True
    else:
        return False



import pandas
from pandas import testing

from conjurer import eda


def test_get_columns_in_dfs():
    input_dfs = [
        pandas.DataFrame(
            data={
                "column1": pandas.Series([], dtype=object),
                "column2": pandas.Series([], dtype="Int64"),
                "column3": pandas.Series([], dtype=float)
            },
            columns=["column1", "column2", "column3"]
        ),
        pandas.DataFrame(
            data={
                "column2": pandas.Series([], dtype="Int64"),
                "column3": pandas.Series([], dtype=object)
            },
            columns=["column2", "column3"],
        ),
        pandas.DataFrame(
            data={
                "column3": pandas.Series([], dtype=object)
            },
            columns=["column3"]
        )
    ]
    name_list = ["table1", "table2", "table3"]
    schema_summary = eda.get_columns_in_dfs(input_dfs, name_list)
    testing.assert_frame_equal(
        schema_summary,
        pandas.DataFrame({
            "table_name": ["table1", "table1", "table1", "table2", "table2", "table3"],
            "column_name": ["column1", "column2", "column3", "column2", "column3", "column3"],
            "dtype": ["object", "Int64", "float", "Int64", "object", "object"]
        })
    )



from conjurer.logic.eda.load import pandas_csv


class DfDictLoader(object):
    """
    Class to load df dictionary as the same dtypes as previous dictionary
    """
    def __init__(self, df_dict: dict):
        self.dtypes = {}
        for name in df_dict.keys():
            self.dtypes[name] = df_dict[name].dtypes

    def load(self, filepath_dict: dict):
        self._validate(filepath_dict)
        return {
            name: pandas_csv.read_csv(filepath_dict[name], dtype=self.dtypes[name])
            for name in filepath_dict.keys()
        }

    def _validate(self, filepath_dict: dict):
        assert set(filepath_dict.keys()) == set(self.dtypes.keys())



import logging

import numpy
import pandas
from IPython.core.display import display
import altair as alt


logger = logging.getLogger(__name__)
METRIC_MINIMIZE = False  # sklearn.metrics.SCORERS are to be maximized
MELT_VARNAME = "data"
HP_INDEX = "hyper parameters index"


class CVAnalyzer(object):
    def __init__(self, pipeline):
        self.param_names = get_param_names(pipeline.cv_results_)
        self.result_df = create_result_df(pipeline.cv_results_, self.param_names)
        self.metric_name = pipeline.scoring
        self._validate()

    def plot_flat(self):
        return plot_flat(self.result_df, self.param_names, self.metric_name)

    def plot_by_param_all(self):
        for param_name in self.param_names:
            self.plot_by_param(param_name).display()

    def plot_by_param(self, param_name):
        return plot_by_param(self.result_df, param_name, self.metric_name, self.param_names)

    def get_summary(self):
        print("Best {} model".format(self.metric_name))
        print(self.get_best_model_info())
        print("Total training time: {} hours for {} params".format(
            self.result_df.fit_time.sum() / 3600, len(self.result_df)))
        display(self.get_param_stat_df())

    def get_best_model_info(self):
        best_model_row = self.result_df.sort_values(
                by="validation_score", ascending=METRIC_MINIMIZE).head(1)
        return dict(
            training=best_model_row.training_score.values[0],
            validation=best_model_row.validation_score.values[0],
            parameters={
                param_name: best_model_row["param_{}".format(param_name)].values[0]
                for param_name in self.param_names
            }
        )

    def get_param_stat_df(self):
        return pandas.concat(
            [get_parameter_wise_stat(self.result_df, param_name)
             for param_name in self.param_names], axis=0).sort_values(
                     by="std (validation score)", ascending=False)

    def _validate(self):
        def _warning(is_min):
            val = self.result_df["param_{}".format(param_name)].min() if is_min \
                else self.result_df["param_{}".format(param_name)].max()
            min_or_max = "minimum" if is_min else "maximum"
            if val == info_dict["parameters"][param_name]:
                logger.warning(
                    "Best performance is achieved at {} value of search: {}={}".format(
                        min_or_max, param_name, info_dict["parameters"][param_name]
                    )
                )

        info_dict = self.get_best_model_info()
        for param_name in self.param_names:
            if _is_numeric(info_dict["parameters"][param_name]):
                _warning(True)
                _warning(False)

        
def create_result_df(cv_result, param_names):
    values_dict = {
        "param_{}".format(key): [
            _transform_param_values_for_groupby(elem[key]) 
            for elem in cv_result["params"]
        ]
        for key in param_names
    }
    values_dict["training_score"] = cv_result["mean_train_score"]
    values_dict["validation_score"] = cv_result["mean_test_score"]
    values_dict["fit_time"] = cv_result["mean_fit_time"]
    values_dict["std_training_score"] = cv_result["std_train_score"]
    values_dict["std_validation_score"] = cv_result["std_test_score"]
    values_dict["std_fit_time"] = cv_result["std_fit_time"]
    return pandas.DataFrame(
        values_dict, 
        columns=["param_{}".format(key) for key in param_names] \
                + ["training_score", "validation_score", "fit_time",
                    "std_training_score", "std_validation_score", "std_fit_time"]
    )


def plot_flat(result_df, param_names, metric_name):
    melt_df = create_melt_df(result_df, param_names)
    return plot_metric_and_time(melt_df, HP_INDEX, metric_name, param_names)


def plot_metric_and_time(melt_df, column_x, metric_name, param_names):
    points = alt.Chart(melt_df[melt_df[MELT_VARNAME].isin(["training_score", "validation_score"])]).encode(
        x=column_x, y=alt.Y("value", title=metric_name, scale=alt.Scale(zero=False)), color=MELT_VARNAME
    ).mark_circle(opacity=0.4)
    mean_df = melt_df.groupby(by=[MELT_VARNAME, column_x]).mean().reset_index().sort_values(by=column_x)
    mean_chart = alt.Chart(
        mean_df[mean_df[MELT_VARNAME].isin(["training_score", "validation_score"])]).encode(
        x=column_x, y=alt.Y("value", title=metric_name, scale=alt.Scale(zero=False)),
        yError="std_value", color=MELT_VARNAME)
    lines = mean_chart.mark_line(strokeDash=[2, 2])
    error = mean_chart.mark_errorband()
    time = alt.Chart(melt_df[melt_df[MELT_VARNAME].isin(["fit_time"])]).mark_bar().encode(
        x=column_x, y=alt.Y("value", title="fit time (sec)"),
        yError="std_value", tooltip=["param_{}".format(param_name) for param_name in param_names]
    )
    return alt.vconcat(
        alt.layer(points, lines+error).properties(height=150, width=400),
        time.properties(height=150, width=400)
    ).resolve_scale(x="shared").properties(title="Accuracy & Computation time")


def plot_by_param(result_df, param_name, metric_name, param_names):
    melt_df = create_melt_df(result_df, param_names)
    return plot_metric_and_time(melt_df, "param_{}".format(param_name), metric_name, param_names)


def create_melt_df(result_df, param_names):
    plot_df = result_df.sort_values(by="validation_score")
    plot_df[HP_INDEX] = range(len(plot_df))
    plot_df[HP_INDEX] = range(len(plot_df))
    melt_df1 = plot_df.melt([HP_INDEX] + ["param_{}".format(param_name) for param_name in param_names],
                            ["training_score", "validation_score", "fit_time"], var_name=MELT_VARNAME)
    melt_df2 = plot_df.melt([HP_INDEX], ["std_training_score", "std_validation_score", "std_fit_time"], 
                            var_name=MELT_VARNAME).rename(
        columns={"value": "std_value"}
    )
    melt_df2[MELT_VARNAME] = melt_df2[MELT_VARNAME].str.replace("std_", "")
    return melt_df1.merge(melt_df2, on=[HP_INDEX, MELT_VARNAME])


def get_parameter_wise_stat(result_df, param_name):
    param_column_name = "param_{}".format(param_name)
    param_values = numpy.sort(result_df[param_column_name].unique())
    mean_metrics = result_df[[param_column_name, "training_score", "validation_score", "fit_time"]].groupby(by=param_column_name).mean()
    return pandas.DataFrame({
            "name": [param_name],
            "uniqueParamValues": [len(param_values)],
            "std (training score)": [numpy.std(mean_metrics.training_score)],
            "std (validation score)": [numpy.std(mean_metrics.validation_score)],
            "std (computation time)": [numpy.std(mean_metrics.fit_time)]
    }).set_index("name")


def get_param_names(cv_results):
    param_names = set([])
    for elem in cv_results["params"]:
        param_names = param_names.union(set(elem.keys()))
    return list(param_names)


def _transform_param_values_for_groupby(value):
    if value is None:
        return "None"
    elif isinstance(value, list):
        return str(value)
    else:
        return value


def _is_numeric(obj):
    return True if obj is not None and numpy.issubdtype(numpy.array([obj]).dtype, numpy.number) else False



import itertools
import copy

import numpy
import pandas
from numpy import random

COLUMN_NEW_ID = "__new_column_for_composite_keys__"


def transform_columns_to_rows(df, column_group, suffixs):
    converted_columns = list(itertools.chain(*[["{}{}".format(c, s) for c in column_group] for s in suffixs]))
    base_columns = [c for c in df.columns if c not in converted_columns]
    ret_df = pandas.concat([
        df[base_columns + ["{}{}".format(c, s) for c in column_group]].rename(
            columns={
                "{}{}".format(c, s): c
                for c in column_group
            }
        )
        for s in suffixs
    ])
    return ret_df.dropna(how="all", subset=column_group)


def random_split_by_ids(df, id_columns, training_ratio):
    ret_df = _add_unique_id(df, id_columns)
    num_new_ids = ret_df[COLUMN_NEW_ID].max() + 1
    training_indices, validation_indices = _get_split_index(num_new_ids, training_ratio)
    return ret_df[ret_df[COLUMN_NEW_ID].isin(training_indices)].drop(columns=[COLUMN_NEW_ID]), \
           ret_df[ret_df[COLUMN_NEW_ID].isin(validation_indices)].drop(columns=[COLUMN_NEW_ID])


def transform_cl_target(series, positive_values, negative_values, positive_label=1, negative_label=0):
    ret_series = pandas.Series(data=[numpy.NaN]*len(series), index=series.index, dtype=series.dtype)
    ret_series[series.isin(positive_values)] = positive_label
    ret_series[series.isin(negative_values)] = negative_label
    return ret_series


def _add_unique_id(df, id_columns):
    id_values = list(set(zip(*[df[c] for c in id_columns])))
    newid_dict = {
        v: i
        for i, v in enumerate(id_values)
    }
    ret_df = copy.deepcopy(df)
    ret_df[COLUMN_NEW_ID] = [newid_dict[v] for v in zip(*[df[c] for c in id_columns])]
    return ret_df


def _get_split_index(num_new_ids, training_ratio):
    indices = list(range(num_new_ids))
    random.shuffle(indices)
    num_training = int(training_ratio * num_new_ids)
    return indices[:num_training], indices[num_training:]



import io

import pandas
from pandas import testing

from conjurer import eda
from . import csv_data


def test_success_execution():
    df = eda.read_csv(io.StringIO(csv_data.ALL_TYPE_TEST_CSV))
    stat_df = eda.check_stats(df)


def test_string():
    df = pandas.DataFrame({
        "str": pandas.Series([pandas.NA, pandas.NA, "aaa", "bbb"])
    })
    stat_df = eda.check_stats(df, True)
    testing.assert_frame_equal(
        stat_df, pandas.DataFrame(
            data={
                "column_name": ["str"],
                "dtype": ["object"],
                "min": [pandas.NA],
                "max": [pandas.NA],
                "mean": [pandas.NA],
                "std": [pandas.NA],
                "ratio_na": [0.5],
                "ratio_zero": [pandas.NA],
                "unique_count": [2],
                "is_unique": [False]
            }
        ))



from scipy import stats
import gbm_autosplit
from . import base


ESTIMATOR_PARAMS = dict(
    boosting_type="gbdt",
    max_depth=-1,
    max_n_estimators=5000,
    early_stopping_rounds=100,
    ratio_training=0.9,
    importance_type="gain"
)
estimator = {
    "cl": gbm_autosplit.LGBMClassifier(metric="auc", **ESTIMATOR_PARAMS),
    "rg": gbm_autosplit.LGBMRegressor(metric="rmse", **ESTIMATOR_PARAMS),
    "mcl": gbm_autosplit.LGBMClassifier(metric="multi_logloss", **ESTIMATOR_PARAMS)
}

grid_rg = dict(
    num_leaves=[10, 20, 40],
    colsample_bytree=[0.1, 0.5],
    learning_rate=[0.01, 0.05, 0.1],
    min_child_samples=[10, 20, 100]
)
grid_cl = {
    **grid_rg,
    **dict(class_weight=["balanced", None])
}
distributions_rg = dict(
    num_leaves=list(range(10, 101)),
    colsample_bytree=stats.uniform(0.05, 0.95),
    learning_rate=stats.uniform(0.01, 0.5),
    min_child_samples=list(range(101))
)
distributions_cl = {
    **distributions_rg,
    **dict(class_weight=["balanced", None])
}
config = base.TunerConfig(estimator, distributions_rg, distributions_cl, grid_rg, grid_cl)



import unittest

from scipy import stats
from sklearn import (
    linear_model,
    tree,
    pipeline,
    impute,
    preprocessing
)

from conjurer import ml
from tests.ml import utils


class TestPandasCV(unittest.TestCase):
    def test_random_linear_holdout_ratio_cl(self):
        model_type, is_cl, with_prep, cv_type = "linear", True, False, "random"
        cv = self._get_cv(model_type, is_cl, with_prep, cv_type)
        self._test_basic_flow_holdout_ratio(cv, is_cl)

    def test_random_tree_holdout_2dfs_rg(self):
        model_type, is_cl, with_prep, cv_type = "tree", False, False, "random"
        cv = self._get_cv(model_type, is_cl, with_prep, cv_type)
        self._test_basic_flow_holdout_2dfs(cv, is_cl)

    def test_random_tree_with_prep_cv_cl(self):
        model_type, is_cl, with_prep, cv_type = "tree", True, True, "random"
        cv = self._get_cv(model_type, is_cl, with_prep, cv_type)
        self._test_basic_flow_cv(cv, is_cl, 3)

    def test_grid_linear_holdout_ratio_cl(self):
        model_type, is_cl, with_prep, cv_type = "linear", True, False, "grid"
        cv = self._get_cv(model_type, is_cl, with_prep, cv_type)
        self._test_basic_flow_holdout_ratio(cv, is_cl)

    def test_grid_tree_holdout_2dfs_rg(self):
        model_type, is_cl, with_prep, cv_type = "tree", False, False, "grid"
        cv = self._get_cv(model_type, is_cl, with_prep, cv_type)
        self._test_basic_flow_holdout_2dfs(cv, is_cl)

    def test_grid_tree_with_prep_cv_cl(self):
        model_type, is_cl, with_prep, cv_type = "tree", True, True, "grid"
        cv = self._get_cv(model_type, is_cl, with_prep, cv_type)
        self._test_basic_flow_cv(cv, is_cl, 3)

    def _get_cv(self, model_type, is_cl, with_prep, cv_type):
        estimator = self._get_estimator(model_type, is_cl, with_prep)
        metric = "roc_auc" if is_cl else "neg_root_mean_squared_error"
        if cv_type == "random":
            params = self._get_params_random(model_type, is_cl, with_prep)
            return ml.RandomizedSearchCV(estimator, params, scoring=metric)
        else:
            params = self._get_params_grid(model_type, is_cl, with_prep)
            return ml.GridSearchCV(estimator, params, scoring=metric)

    def _get_estimator(self, model_type, is_cl, with_preprocessing):
        if model_type == "linear":
            ml_estimator = linear_model.LogisticRegression(solver="liblinear") if is_cl else linear_model.Lasso()
        else:
            ml_estimator = tree.DecisionTreeClassifier() if is_cl else tree.DecisionTreeRegressor()
        return self._add_preprocessing(ml_estimator) if with_preprocessing else ml_estimator

    @staticmethod
    def _add_preprocessing(estimator):
        return pipeline.Pipeline(
            steps=[
                ("mvi", impute.SimpleImputer()),
                ("std", preprocessing.StandardScaler()),
                ("ml", estimator)
            ]
        )

    def _get_params_random(self, model_type, is_cl, with_preprocessing):
        if model_type == "linear":
            ml_params = dict(
                penalty=["l1", "l2"],
                C=stats.loguniform(1e-5, 10)
            ) if is_cl else dict(alpha=stats.loguniform(1e-5, 10))
        else:
            ml_params = dict(max_depth=list(range(5, 16)))
        return self._convert_ml_params(ml_params) if with_preprocessing else ml_params

    def _get_params_grid(self, model_type, is_cl, with_preprocessing):
        if model_type == "linear":
            ml_params = dict(
                penalty=["l1", "l2"],
                C=[1e-5, 1e-3]
            ) if is_cl else dict(alpha=[1e-5, 1e-3, 1e-1, 10])
        else:
            ml_params = dict(max_depth=[5, 8, 11, 14])
        return self._convert_ml_params(ml_params) if with_preprocessing else ml_params

    @staticmethod
    def _convert_ml_params(ml_params):
        return {"{}__{}".format("ml", k): v for k, v in ml_params.items()}

    def _test_basic_flow_holdout_ratio(self, cv, is_cl):
        df_training = utils.get_input_df(100)
        df_test = utils.get_input_df(10)
        target_column = "target_cl" if is_cl else "target_rg"
        feature_columns = ["column{}".format(i) for i in range(6)]
        model = cv.fit_holdout_pandas(df_training, target_column, feature_columns, ratio_training=0.8)
        self._assert_prediction(model, df_test, is_cl)

    def _test_basic_flow_holdout_2dfs(self, cv, is_cl):
        df_training = utils.get_input_df(100)
        df_validation = utils.get_input_df(100)
        df_test = utils.get_input_df(10)
        target_column = "target_cl" if is_cl else "target_rg"
        feature_columns = ["column{}".format(i) for i in range(6)]
        model = cv.fit_holdout_pandas(df_training, target_column, feature_columns, df_validation)
        self._assert_prediction(model, df_test, is_cl)

    def _test_basic_flow_cv(self, cv, is_cl, n_fold):
        df_training = utils.get_input_df(100)
        df_test = utils.get_input_df(10)
        target_column = "target_cl" if is_cl else "target_rg"
        feature_columns = ["column{}".format(i) for i in range(6)]
        model = cv.fit_cv_pandas(df_training, target_column, feature_columns, n_fold=n_fold)
        self._assert_prediction(model, df_test, is_cl)

    def _assert_prediction(self, model, df_test, is_cl):
        pred_df = model.predict(df_test)
        expected_columns = ["score", "id1", "id2", "target_cl", "target_rg"]
        if is_cl:
            expected_columns.insert(1, "predicted_class")
        self.assertListEqual(list(pred_df.columns), expected_columns)
        self.assertEqual(len(pred_df), 10)



from conjurer import ml
from tests.ml import utils


def test_lm_cv_grid():
    cv = ml.get_default_cv("linear_model", "cl", search_type="grid")
    cv.param_grid = {
        "ml__penalty": ["l1"],
        "ml__C": [1e-5, 1e-3, 1e-1]
    }
    _test_basic_flow_holdout_pandas1(cv, True)
    _test_basic_flow_holdout_pandas2(cv, True)
    _test_basic_flow_cv_pandas(cv, True)


def test_lm_cv_random():
    cv = ml.get_default_cv("linear_model", "cl", search_type="random")
    cv.n_iter = 2
    _test_basic_flow_holdout_pandas1(cv, True)
    _test_basic_flow_holdout_pandas2(cv, True)
    _test_basic_flow_cv_pandas(cv, True)


def test_lightgbm_cv_grid():
    cv = ml.get_default_cv("lightgbm", "cl", search_type="grid")
    cv.param_grid = {
        "ml__num_leaves": [10],
        "ml__colsample_bytree": [0.1],
        "ml__learning_rate": [0.01],
        "ml__min_child_samples": [0, 20, 100]
    }
    _test_basic_flow_holdout_pandas1(cv, True)
    _test_basic_flow_holdout_pandas2(cv, True)
    _test_basic_flow_cv_pandas(cv, True)


def test_lightgbm_cv_random():
    cv = ml.get_default_cv("lightgbm", "cl", search_type="random")
    cv.n_iter = 5
    _test_basic_flow_holdout_pandas1(cv, True)
    _test_basic_flow_holdout_pandas2(cv, True)
    _test_basic_flow_cv_pandas(cv, True)


def test_lightgbm_cv_random_rg():
    cv = ml.get_default_cv("lightgbm", "rg", search_type="random")
    cv.n_iter = 5
    _test_basic_flow_holdout_pandas1(cv, False)
    _test_basic_flow_holdout_pandas2(cv, False)
    _test_basic_flow_cv_pandas(cv, False)


def test_xgboost_cv_grid():
    cv = ml.get_default_cv("xgboost", "rg", search_type="grid")
    cv.param_grid = dict(
        ml__colsample_bynode=[0.1],
        ml__learning_rate=[0.01],
        ml__ratio_min_child_weight=[None, 0.005, 0.01]
    )
    _test_basic_flow_holdout_pandas1(cv, False)
    _test_basic_flow_holdout_pandas2(cv, False)
    _test_basic_flow_cv_pandas(cv, False)


def test_xgboost_cv_random():
    cv = ml.get_default_cv("xgboost", "cl", search_type="random")
    cv.n_iter = 5
    _test_basic_flow_holdout_pandas1(cv, True)
    _test_basic_flow_holdout_pandas2(cv, True)
    _test_basic_flow_cv_pandas(cv, True)


def test_random_forest_cv_grid():
    cv = ml.get_default_cv("random_forest", "cl", search_type="grid")
    cv.param_grid = dict(
        ml__max_depth=[2],
        ml__n_estimators=[100],
        ml__max_features=["auto"],
        ml__min_samples_leaf=[0.01, 0.05, 0.1]
    )
    _test_basic_flow_holdout_pandas1(cv, True)
    _test_basic_flow_holdout_pandas2(cv, True)
    _test_basic_flow_cv_pandas(cv, True)


def test_random_forest_cv_random():
    cv = ml.get_default_cv("random_forest", "rg", search_type="random")
    cv.n_iter = 5
    _test_basic_flow_holdout_pandas1(cv, False)
    _test_basic_flow_holdout_pandas2(cv, False)
    _test_basic_flow_cv_pandas(cv, False)


def _test_basic_flow_holdout_pandas1(cv, is_cl):
    df_training = utils.get_input_df(100)
    df_test = utils.get_input_df(10)
    target_column = "target_cl" if is_cl else "target_rg"
    feature_columns = ["column{}".format(i) for i in range(6)]
    model = cv.fit_holdout_pandas(df_training, target_column, feature_columns, ratio_training=0.8)
    _assert_prediction(model, df_test, is_cl)


def _test_basic_flow_holdout_pandas2(cv, is_cl):
    df_training = utils.get_input_df(100)
    df_validation = utils.get_input_df(100)
    df_test = utils.get_input_df(10)
    target_column = "target_cl" if is_cl else "target_rg"
    feature_columns = ["column{}".format(i) for i in range(6)]
    model = cv.fit_holdout_pandas(df_training, target_column, feature_columns, df_validation)
    _assert_prediction(model, df_test, is_cl)


def _test_basic_flow_cv_pandas(cv, is_cl):
    df_training = utils.get_input_df(100)
    df_test = utils.get_input_df(10)
    target_column = "target_cl" if is_cl else "target_rg"
    feature_columns = ["column{}".format(i) for i in range(6)]
    model = cv.fit_cv_pandas(df_training, target_column, feature_columns, 3)
    _assert_prediction(model, df_test, is_cl)


def _assert_prediction(model, df_test, is_cl):
    pred_df = model.predict(df_test)
    expected_columns = ["score", "id1", "id2", "target_cl", "target_rg"]
    if is_cl:
        expected_columns.insert(1, "predicted_class")
    assert list(pred_df.columns) == expected_columns
    assert len(pred_df) == 10
